
import numpy as np

## Example from Salman Durrani's course notes:
#
#   x1 +   x2 + x3 = 5
#  -x1 + 2 x2      = 9
# 4 x1 +   x2 - x3 = -2

A1 = np.array([[ 1, 1, 1],
               [-1, 2, 0],
               [ 4, 1,-1]])
b1 = np.array([5, 9, -2])

## Example from Lab 2, Exercise 3:
#
# SA + 2 CNK + 3.8 CWK + 2.8 SWK = 196037
#        CNK +     CWK +     SWK = 50352
#                  CWK           = 22850
#                            SWK = 7243
#
# (SA: single adults; CNK: couples no kids; CWK: couples
# with kids; SWK: single parents with kids)

A2 = np.array([[ 1, 2, 3.8, 2.8],
               [ 0, 1, 1  , 1  ],
               [ 0, 0, 1  , 0  ],
               [ 0, 0, 0  , 1  ]])
b2 = np.array([196037, 50352, 22850, 7243])


def solve_cramer(A, b):
    '''Solve the system of equations Ax=b using Cramer's rule and
    return x.
    Assumptions: A is an n*n coefficient matrix, b is a n*1 vector.'''
    n = len(b)
    # Compute the determinant of A:
    detA = np.linalg.det(A)
    # Create an array to store the solution; the initial values in
    # the array do not matter because they will all be replaced in
    # the loop below.
    x = np.zeros(n)
    for i in np.arange(n):
        # For each unknown x_i:
        # make a copy of the i:th column of A
        tmp = np.copy(A[:,i])
        # replace i:th column of A with b
        A[:,i] = b
        # compute the determinant of the modified matrix:
        detAi = np.linalg.det(A)
        # put back the original i:th column
        A[:,i] = tmp
        # compute x_i using Cramer's rule:
        x[i] = detAi / detA
    return x


def solve_gauss(A, b):
    '''Solve the system of equations Ax=b using Gauss' elimination
    and return x.
    Assumptions: A is an n*n coefficient matrix, b is a n*1 vector.'''
    n = len(b)
    # Construct the extended matrix, E, by concatenating A and b
    E = np.concatenate((A,np.reshape(b, (n,1))), axis=1)
    # Put E into triangular form:
    # For each column, j, except the last:
    for j in np.arange(n - 1):
        # for each row i below row j:
        for i in np.arange(j + 1, n):
            # find the ratio, r, between the coefficients E[i, j]
            # and E[j, j], and subtract r * row j from row i. The
            # result of this is that all rows below j will now have
            # a zero coefficient in column j.
            r = E[i, j] / E[j, j]
            E[i,:] = E[i,:] - r * E[j,:]
    # Create an array to store the result
    x = np.zeros(n)
    # Now, we can solve the equations in reverse order, i.e., starting
    # with x_n and working back to x_1.
    for i in (n - 1) - np.arange(n):
        # To find x_i, take the right-hand side value from the extended
        # matrix, E[i,n], and subtract the values of x_i+1 .. x_n multiplied
        # by their coefficients in row i, and divide the result by the
        # coefficient of x_i:
        x[i] = (E[i,n] - np.dot(E[i,(i+1):n], x[(i+1):n])) / E[i,i]
    return x


def test_eq_sys(n):
    '''Return a random n*n system of linear equations.
    This is done by generating an n*n matrix of random numbers (in the
    range [0,1)) and adding up the columns to make the right-hand side.
    The solution to this system should be x_1 = 1, .., x_n = 1, i.e.,
    a row of all ones.'''
    A = np.random.rand(n, n)
    b = np.zeros(n)
    for col in np.arange(n):
        b = b + A[:,col]
    return A, b
